'''
    TweeLoc is created by

    Pavlos Paraskevopoulos:
        George Mason University
        LIPADE, Paris Descartes University
        pparaske@gmu.edu

    Giovanni Pellegrini:
        University of Trento
        giovanni.pellegrini@unitn.it

    Themis Palpanas:
        Paris Descartes University
        themis@mi.parisdescartes.fr

	Copyright Paraskevopoulos, Pellegrini, Palpanas.


	This software cannot be used for commercial purposes.

	You may use the software as is, only for research purposes.
	You should acknowledge the authors by referencing the following two papers (the first describes the method and algorithms; the second describes the system and visualizations):

	Pavlos Paraskevopoulos, Themis Palpanas.
	Where has this Tweet Come From? Fast and Fine-Grained Geolocalisation of Non-Geotagged Tweets. Social Network Analysis and Mining (SNAM) Journal 6(1), pp 1-16, 2016.
	http://www.mi.parisdescartes.fr/~themisp/publications/snam16.pdf

	Pavlos Paraskevopoulos, Giovanni Pellegrini, Themis Palpanas. TweeLoc: A System for Geolocalizing Tweets at Fine-Grain. International Conference on Data Mining (ICDM), New Orleans, LA, USA, November 2017.
	http://www.mi.parisdescartes.fr/~themisp/publications/icdm17-tweeloc.pdf
'''

import preprocessing
import geolocalisation
import time
from datetime import timedelta
from threading import Thread


class TweeLocAlgorithmThread(Thread):
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs=None, verbose=None, algorithm=None, configuration=None):
        Thread.__init__(self, group=group, target=target, name=name, verbose=verbose)
        self.algorithm = algorithm
        self.conf = configuration
        return

    def run(self):
        if isinstance(self.algorithm, TweeLocAlgorithmOffline):
            start_preprocessing_time = time.time()
            print "===== Execution started at : {0} =====".format(
                time.strftime("%a %b %d %H:%M:%S %Y", time.localtime(time.time())))
            try:
                self.algorithm.preprocessing()
            except Exception as e:
                print "Error in the preprocessing phase: {0}. Program will be closed.".format(e.__str__())
                self.algorithm.work_done = True
                return

            stop_preprocessing_time = time.time()
            m, s = divmod(stop_preprocessing_time - start_preprocessing_time, 60)
            print "===== Preprocessing time: {0} m  {1} s =====".format(int(m), int(s))

            start_analyzing_time = time.time()
            try:
                self.algorithm.analyzing()
            except Exception as e:
                print "Error in the analyzing phase: {0}. Program will be closed.".format(e.__str__())
                raise
            stop_analyzing_time = time.time()
            m, s = divmod(stop_analyzing_time - start_analyzing_time, 60)
            print "\n"
            print "===== Analyzing time: {0} m  {1} s =====".format(int(m), int(s))
            self.merged_results = preprocessing.merge_results(self.conf)
            print "Creating plots of the performance..."
        # plots.create_plots_performance(merged_results, self.conf)

        elif isinstance(self.algorithm, TweeLocAlgorithmOnline):
            start_time = time.time()
            print "===== Execution started at : {0} =====".format(
                time.strftime("%a %b %d %H:%M:%S %Y", time.localtime(time.time())))
            try:
                self.algorithm.preprocessing()
            except Exception as e:
                print "Error in the analyzing phase: {0}. Program will be closed.".format(e.__str__())
                raise
            stop_time = time.time()
            m, s = divmod(stop_time - start_time, 60)
            print "===== Execution time: {0} m  {1} s =====".format(int(m), int(s))
            self.merged_results = preprocessing.merge_results(self.conf)
            print "Creating plots of the performance..."
        # plots.create_plots_performance(merged_results, self.conf)
        return

    def get_results(self):
        return self.merged_results


class TweeLocBase:
    def __init__(self):
        self.conf = None
        self.stop = False

    def _group_tweets(self, location, coordinates):
        print "Grouping {0} tweets by location and squares...".format(location.name)
        # assign to each square the tweets done in that area
        tweets_sq = preprocessing.group_tweets_by_squares(self.tweets, coordinates[0], coordinates[1], self.conf.mask)
        print "Grouping {0} tweets by time slots...\n".format(location.name)
        # for each square and timeslot, assign tweets done in that place at that moment
        tweets_sq_ts = preprocessing.group_tweets_by_timeslot(tweets_sq, self.time_slots_id)
        return tweets_sq_ts

    def _create_locations_square_grid(self):
        self.locations_squares = {}
        for location in self.conf.locations:
            print "Creating {0} squares...".format(location.name)
            # divide the input area of a city into squares, and return the position of the squares and their indexes
            cols_coordinates_id, rows_coordinates_id = preprocessing.create_squares_id(
                {'lat': location.lat, 'lon': location.lon}, location.x_side, location.y_side, self.conf.l2_side)
            self.locations_squares[location.name] = (cols_coordinates_id, rows_coordinates_id)
            preprocessing.save_location_coordinates(self.conf, location.name,
                                                    (cols_coordinates_id, rows_coordinates_id))
        print ""
        return self.locations_squares

    def _iterate_keywords_percentage(self, start_ts, suffix, train, test, n_test, cycle):
        for keyword_percentage in sorted(self.conf.keywords_percentage):
            if keyword_percentage == max(self.conf.keywords_percentage) and cycle == 0 and (
                    start_ts >= self.conf.n_ts_per_day or self.conf.thresholds_structure):
                save = True
            else:
                save = False
            if keyword_percentage == max(self.conf.keywords_percentage) and self.conf.stream and cycle == 0 and (
                    start_ts >= self.conf.n_ts_per_day or self.conf.thresholds_structure):
                stream = True
            else:
                stream = False
            geolocalisation.print_progress(start_ts + 1, max(self.time_slots_id.keys()) - self.conf.window + 2,
                                           prefix='Progress:',
                                           suffix=suffix + " | {0}% Keywords".format(keyword_percentage), barLength=50)

            # each city or square is a document, each of its word is associate with its tfidf value
            documents_l1, idf_l1 = geolocalisation.create_tfidf_documents(train, 'L1')

            # remove words with less tfidf value, according to the percentage value
            pruned_documents_l1 = geolocalisation.prune_documents(documents_l1, keyword_percentage)

            self.prediction.predictions_l1 = geolocalisation.predict_tweets_l1_loc(self.conf, test, pruned_documents_l1,
                                                                                   idf_l1, self.thresholds,
                                                                                   keyword_percentage, start_ts,
                                                                                   self.l1_correlations)

            if stream:
                self.prediction.predictions_l1_position = geolocalisation.predict_tweets_l1_no_position(self.predict,
                                                                                                        pruned_documents_l1,
                                                                                                        idf_l1,
                                                                                                        self.thresholds,
                                                                                                        keyword_percentage,
                                                                                                        start_ts,
                                                                                                        self.l1_correlations,
                                                                                                        self.conf.n_ts_per_day)

            similarities_l2 = {}
            similarities_l2_position = {}
            for t in self.conf.thresholds:
                similarities_l2[t] = {}

            for l1_id, l2_list in train.iteritems():
                documents_l2, idf_l2 = geolocalisation.create_tfidf_documents(l2_list, 'L2')
                pruned_documents_l2 = geolocalisation.prune_documents(documents_l2, keyword_percentage)

                similarities_l2 = geolocalisation.predict_tweets_l2_loc(self.conf, test, pruned_documents_l2, idf_l2,
                                                                        self.thresholds, keyword_percentage, start_ts,
                                                                        self.l2_correlations, l1_id, similarities_l2)

                if stream:
                    similarities_l2_position = geolocalisation.predict_tweets_l2_no_position(self.predict,
                                                                                             pruned_documents_l2,
                                                                                             idf_l2, self.thresholds,
                                                                                             keyword_percentage,
                                                                                             start_ts,
                                                                                             self.l2_correlations,
                                                                                             self.conf.n_ts_per_day,
                                                                                             similarities_l2_position,
                                                                                             l1_id)

                if save:
                    preprocessing.save_square_keywords(l1_id, documents_l2, self.time_slots_id[start_ts]['start'],
                                                       self.conf)

            self.prediction.predictions_l2 = similarities_l2
            self.prediction.predictions_l2_position = similarities_l2_position

            if start_ts < self.conf.n_ts_per_day and not self.conf.thresholds_structure:
                probabilities = geolocalisation.merge_probabilities(self.prediction.predictions_l1[0],
                                                                    self.prediction.predictions_l2[0])
                top_matches = geolocalisation.extract_most_probable_place(probabilities)
                self.thresholds = geolocalisation.update_thresholds(self.conf, self.thresholds, test, top_matches,
                                                                    self.prediction.predictions_l1[0],
                                                                    self.prediction.predictions_l2[0],
                                                                    keyword_percentage, start_ts)
            else:
                if stream:
                    predict_probabilities = geolocalisation.merge_probabilities(self.prediction.predictions_l1_position,
                                                                                self.prediction.predictions_l2_position)
                    top_matches_prediction = geolocalisation.extract_most_probable_place(predict_probabilities)
                    preprocessing.save_prediction(self.predict, top_matches_prediction, self.locations_squares,
                                                  self.time_slots_id[start_ts]['start'], self.conf)

                for t in self.conf.thresholds:
                    probabilities = geolocalisation.merge_probabilities(self.prediction.predictions_l1[t],
                                                                        self.prediction.predictions_l2[t])
                    top_matches = geolocalisation.extract_most_probable_place(probabilities)
                    precision, recall, f1 = geolocalisation.confront_prediction_actual_test(test, top_matches, n_test)
                    if precision != -1 and recall != -1:
                        self.prediction.results[keyword_percentage][t]['precision'].append(precision)
                        self.prediction.results[keyword_percentage][t]['recall'].append(recall)
                        self.prediction.results[keyword_percentage][t]['f1'].append(f1)

                    if t == 0:
                        self.thresholds = geolocalisation.update_thresholds(self.conf, self.thresholds, test,
                                                                            top_matches,
                                                                            self.prediction.predictions_l1[0],
                                                                            self.prediction.predictions_l2[0],
                                                                            keyword_percentage, start_ts)

                        if save:
                            save = False
                            preprocessing.save_train_data(train, self.time_slots_id[start_ts]['start'], self.conf)
                            preprocessing.save_tweets_not_location(self.tweets_not_in_location,
                                                                   self.time_slots_id[start_ts]['start'], start_ts,
                                                                   self.conf)
                            preprocessing.save_test_data(test, top_matches, self.locations_squares,
                                                         self.time_slots_id[start_ts]['start'], self.conf)
                            preprocessing.save_square_train_test_counter(train, test, top_matches,
                                                                         self.time_slots_id[start_ts]['start'],
                                                                         self.conf)
                            preprocessing.save_square_differential(self.sq_counter_curr, self.sq_counter_prev,
                                                                   self.time_slots_id[start_ts]['start'], self.conf)


class TweeLocAlgorithmOffline(TweeLocBase):
    def __init__(self, conf):
        self.conf = conf
        self.stop = False
        self.work_done = False

    def preprocessing(self):

        self.tweets = preprocessing.retrieve_tweets_from_folder(self.conf)
        # Keep only tweets within a certain range of time, from start_time to stop_time
        self.tweets, self.tweets_to_predict = preprocessing.extract_tweets_start_time_stop_time(self.tweets,
                                                                                                self.conf.start_dt,
                                                                                                self.conf.stop_dt)
        if len(self.tweets.keys()) == 0:
            raise Exception("No useful tweets found in the files, please check your tweets")
        print "\n"
        print "Number of tweets in the range {0} - {1} : {2}".format(self.conf.start_dt, self.conf.stop_dt,
                                                                     (len(self.tweets) + len(self.tweets_to_predict)))
        print "Tweets with coordinates: {0}".format(len(self.tweets))
        print "Tweets without coordinates: {0}\n".format((len(self.tweets_to_predict)))
        print "Creating time slots...\n"

        # divide the time serie into slots and it assigns to each slot an id
        self.time_slots_id = preprocessing.create_time_slots_id(self.conf.start_dt, self.conf.stop_dt,
                                                                self.conf.timeslot_size)
        tweets_l1_locations = {}
        self.locations_squares = {}
        self.tweets_not_in_location = {}
        self._create_locations_square_grid()
        # Divide each city into squares and assign each tweet to its square in the city
        for location in self.conf.locations:
            tweets_l1_locations[location.name] = self._group_tweets(location, self.locations_squares[location.name])

        tweets_not_location = preprocessing.extract_tweets_not_in_locations(self.tweets, tweets_l1_locations)
        print "Grouping tweets not in the locations above...\n"
        self.tweets_not_in_location = preprocessing.group_tweets_by_timeslot(tweets_not_location, self.time_slots_id)
        self.tweets = tweets_l1_locations
        del tweets_l1_locations
        del tweets_not_location

        if self.conf.stream:
            tweets_to_predict_ts = preprocessing.group_tweets_to_predict_by_timeslot(self.tweets_to_predict,
                                                                                     self.time_slots_id)
            self.tweets_to_predict = tweets_to_predict_ts
            del tweets_to_predict_ts

        preprocessing.save_squares_time_series(self.tweets, self.time_slots_id, self.conf)

    def analyzing(self):
        for cycle in xrange(0, self.conf.n_runs):
            if self.conf.thresholds_structure:
                self.thresholds = self.conf.thresholds_structure
            else:
                self.thresholds = geolocalisation.create_threshold_structure(self.conf.keywords_percentage,
                                                                             self.conf.n_ts_per_day)
            print "\n"
            print "======================================================================"
            print "Running {0} of {1} cycles".format(cycle + 1, self.conf.n_runs)
            print "1 Day = {0} Time slots".format(self.conf.n_ts_per_day)

            self.prediction = geolocalisation.Prediction(self.conf.thresholds, self.conf.keywords_percentage)
            self.sq_counter_curr = {}

            for start_ts in xrange(0, max(self.time_slots_id.keys()) - self.conf.window + 2, 2):
                if self.stop:
                    break
                suffix = "Time slot #{0} - {1}".format(start_ts, self.time_slots_id[start_ts]['start'])

                if self.conf.stream:
                    self.predict = geolocalisation.get_tweets_to_predict(self.tweets_to_predict, start_ts)

                # Divide the entire set into train set and test set
                train, test, n_test = geolocalisation.split_dataset_train_test(self.tweets, self.conf.train, start_ts,
                                                                               self.conf.window)

                if (start_ts >= self.conf.n_ts_per_day or self.conf.thresholds_structure) and cycle == 0:
                    preprocessing.create_timeslot_folder(self.conf, train.keys(), self.time_slots_id[start_ts]['start'])
                self.sq_counter_curr, self.sq_counter_prev = preprocessing.calculate_differential_heatmap(self.conf,
                                                                                                          start_ts,
                                                                                                          train,
                                                                                                          self.sq_counter_curr)

                # Calculate the correlations of the activities for the cities and the squares of a city
                if self.conf.l1_correlation:
                    # Activities of the cities
                    local_l1_activities, global_l1_activity = geolocalisation.local_global_activities(self.tweets,
                                                                                                      self.conf.window,
                                                                                                      start_ts, 'L1')
                    # Calculate the correlation between the local activities and the global activity of the cities
                    l1_correlations = geolocalisation.correlations(local_l1_activities, global_l1_activity,
                                                                   self.conf.window)
                    self.l1_correlations = l1_correlations
                else:
                    self.l1_correlations = None

                if self.conf.l2_correlation:
                    l2_correlations = {}
                    for l1_id in self.tweets.keys():
                        # Activities of the squares of a city
                        local_l2_activities, global_l2_activity = geolocalisation.local_global_activities(
                            self.tweets[l1_id], self.conf.window, start_ts, 'L2')
                        # Calculate the correlation between the local activities and the global activity of the squares
                        l2_correlations[l1_id] = geolocalisation.correlations(local_l2_activities, global_l2_activity,
                                                                              self.conf.window)
                    self.l2_correlations = l2_correlations
                else:
                    self.l2_correlations = None

                self._iterate_keywords_percentage(start_ts, suffix, train, test, n_test, cycle)
            preprocessing.save_results(self.conf, cycle, self.prediction.results)
            preprocessing.save_thresholds(self.conf, self.thresholds)
        self.work_done = True


class TweeLocAlgorithmOnline(TweeLocBase):
    def __init__(self, conf, downloader):
        self.conf = conf
        self.downloader = downloader
        self.stop = False
        self.work_done = False

    def preprocessing(self):
        file_queue = []

        self.locations_squares = {}
        self._create_locations_square_grid()
        self.thresholds = geolocalisation.create_threshold_structure(self.conf.keywords_percentage,
                                                                     self.conf.n_ts_per_day)
        self.prediction = geolocalisation.Prediction(self.conf.thresholds, self.conf.keywords_percentage)
        self.sq_counter_curr = {}

        start_ts = 0
        first_time = True
        print "1 Day = {0} Time slots".format(self.conf.n_ts_per_day)
        while not self.stop:
            if (self.downloader.file_queue.qsize() + len(file_queue)) < self.conf.window:
                preprocessing.print_window_progress(self.downloader.file_queue.qsize() + len(file_queue),
                                                    self.conf.window)
                time.sleep(1.0)
                continue
            preprocessing.print_window_progress(self.downloader.file_queue.qsize(), self.conf.window)
            print ""
            while not self.downloader.file_queue.empty():
                file_queue.append(self.downloader.file_queue.get())
            file_queue = sorted(file_queue)

            self.tweets, self.tweets_to_predict, start_datetime = preprocessing.retrieve_tweets_online(self.conf,
                                                                                                       file_queue[
                                                                                                       :self.conf.window])
            if first_time:
                first_time = False
            else:
                start_datetime = start_datetime + timedelta(minutes=self.conf.timeslot_size * 2)
            # print len(self.tweets), start_datetime
            print "N. of tweets: {0}".format(len(self.tweets))

            self.time_slots_id = preprocessing.create_time_slots_id_online(start_datetime, start_ts, self.conf)
            preprocessing.print_time_slots_progress(min(self.time_slots_id.keys()))

            tweets_l1_locations = {}
            self.tweets_not_in_location = {}
            # Divide each city into squares and assign each tweet to its square in the city
            for location in self.conf.locations:
                tweets_l1_locations[location.name] = self._group_tweets(location, self.locations_squares[location.name])
            tweets_not_location = preprocessing.extract_tweets_not_in_locations(self.tweets, tweets_l1_locations)
            self.tweets_not_in_location = preprocessing.group_tweets_by_timeslot(tweets_not_location,
                                                                                 self.time_slots_id)
            self.tweets = tweets_l1_locations
            del tweets_l1_locations
            del tweets_not_location

            if self.conf.stream:
                tweets_to_predict_ts = preprocessing.group_tweets_to_predict_by_timeslot(self.tweets_to_predict,
                                                                                         self.time_slots_id)
                self.tweets_to_predict = tweets_to_predict_ts
                del tweets_to_predict_ts
            # preprocessing.save_squares_time_series(self.tweets, self.time_slots_id, self.conf)

            self.analyzing(start_ts)
            print ""
            # print "FILE_QUEUE: ", file_queue
            del file_queue[0]
            del file_queue[0]

            start_ts += 2
            print "==========================="

        preprocessing.save_results(self.conf, 0, self.prediction.results)
        preprocessing.save_thresholds(self.conf, self.thresholds)
        self.work_done = True

    def analyzing(self, start_ts):
        cycle = 0
        suffix = "Time slot #{0} - {1}".format(start_ts, self.time_slots_id[start_ts]['start'])

        if self.conf.stream:
            self.predict = geolocalisation.get_tweets_to_predict(self.tweets_to_predict, start_ts)

        # Divide the entire set into train set and test set
        train, test, n_test = geolocalisation.split_dataset_train_test(self.tweets, self.conf.train, start_ts,
                                                                       self.conf.window)

        if start_ts >= self.conf.n_ts_per_day and cycle == 0:
            preprocessing.create_timeslot_folder(self.conf, train.keys(), self.time_slots_id[start_ts]['start'])

        self.sq_counter_curr, self.sq_counter_prev = preprocessing.calculate_differential_heatmap(self.conf, start_ts,
                                                                                                  train,
                                                                                                  self.sq_counter_curr)

        # Calculate the correlations of the activities for the cities and the squares of a city
        if self.conf.l1_correlation:
            # Activities of the cities
            local_l1_activities, global_l1_activity = geolocalisation.local_global_activities(self.tweets,
                                                                                              self.conf.window,
                                                                                              start_ts, 'L1')
            # Calculate the correlation between the local activities and the global activity of the cities
            l1_correlations = geolocalisation.correlations(local_l1_activities, global_l1_activity, self.conf.window)
            self.l1_correlations = l1_correlations
        else:
            self.l1_correlations = None

        if self.conf.l2_correlation:
            l2_correlations = {}
            for l1_id in self.tweets.keys():
                # Activities of the squares of a city
                local_l2_activities, global_l2_activity = geolocalisation.local_global_activities(self.tweets[l1_id],
                                                                                                  self.conf.window,
                                                                                                  start_ts, 'L2')
                # Calculate the correlation between the local activities and the global activity of the squares
                l2_correlations[l1_id] = geolocalisation.correlations(local_l2_activities, global_l2_activity,
                                                                      self.conf.window)
            self.l2_correlations = l2_correlations
        else:
            self.l2_correlations = None

        self._iterate_keywords_percentage(start_ts, suffix, train, test, n_test, cycle)
