'''
    TweeLoc is created by

    Pavlos Paraskevopoulos:
        George Mason University
        LIPADE, Paris Descartes University
        pparaske@gmu.edu

    Giovanni Pellegrini:
        University of Trento
        giovanni.pellegrini@unitn.it

    Themis Palpanas:
        Paris Descartes University
        themis@mi.parisdescartes.fr

	Copyright Paraskevopoulos, Pellegrini, Palpanas.


	This software cannot be used for commercial purposes.

	You may use the software as is, only for research purposes.
	You should acknowledge the authors by referencing the following two papers (the first describes the method and algorithms; the second describes the system and visualizations):

	Pavlos Paraskevopoulos, Themis Palpanas.
	Where has this Tweet Come From? Fast and Fine-Grained Geolocalisation of Non-Geotagged Tweets. Social Network Analysis and Mining (SNAM) Journal 6(1), pp 1-16, 2016.
	http://www.mi.parisdescartes.fr/~themisp/publications/snam16.pdf

	Pavlos Paraskevopoulos, Giovanni Pellegrini, Themis Palpanas. TweeLoc: A System for Geolocalizing Tweets at Fine-Grain. International Conference on Data Mining (ICDM), New Orleans, LA, USA, November 2017.
	http://www.mi.parisdescartes.fr/~themisp/publications/icdm17-tweeloc.pdf
'''


import numpy as np
import math
from geoplotlib import colors
from geoplotlib.layers import BaseLayer
from geoplotlib.core import BatchPainter
from geoplotlib.utils import BoundingBox


class DensityLayer(BaseLayer):
	def __init__(self, data, bw, cmap='hot', method='hist', scaling='sqrt', alpha=220,
				 cut_below=None, clip_above=None, binsize=1, cmap_levels=10, show_colorbar=False, communicator=None):
		"""
		Kernel density estimation visualization

		:param data: data access object
		:param bw: kernel bandwidth (in screen coordinates)
		:param cmap: colormap
		:param method: if kde use KDEMultivariate from statsmodel, which provides a more accurate but much slower estimation.
			If hist, estimates density applying gaussian smoothing on a 2D histogram, which is much faster but less accurate
		:param scaling: colorscale, lin log or sqrt
		:param alpha: color alpha
		:param cut_below: densities below cut_below are not drawn
		:param clip_above: defines the max value for the colorscale
		:param binsize: size of the bins for hist estimator
		:param cmap_levels: discretize colors into cmap_levels
		:param show_colorbar: show colorbar
		:param communicator: interface to communicate with the app
		"""
		self.data = data
		self.bw = bw
		self.cmap = colors.ColorMap(cmap, alpha=alpha, levels=cmap_levels)
		self.method = method
		self.scaling = scaling
		self.cut_below = cut_below
		self.clip_above = clip_above
		self.binsize = binsize
		self.show_colorbar = show_colorbar
		self.communicator = communicator

	def _get_grid(self, proj):
		west, north = proj.lonlat_to_screen([proj.bbox().west], [proj.bbox().north])
		east, south = proj.lonlat_to_screen([proj.bbox().east], [proj.bbox().south])
		longadd = math.fabs(east - west) * 0.5
		latadd = math.fabs(north - south) * 0.5
		xgrid = np.arange(west - longadd, east + longadd, self.binsize)
		ygrid = np.arange(south - latadd, north + latadd, self.binsize)
		return xgrid, ygrid

	def invalidate(self, proj):
		self.painter = BatchPainter()
		try:
			xv, yv = proj.lonlat_to_screen(self.data['lon'], self.data['lat'])

			rects_vertices = []
			rects_colors = []

			if self.method == 'hist':
				try:
					from scipy.ndimage import gaussian_filter
				except:
					raise Exception('KDE requires scipy')

				xgrid, ygrid = self._get_grid(proj)
				H, _, _ = np.histogram2d(yv, xv, bins=(ygrid, xgrid))

				if H.sum() == 0:
					print('no data in current view')
					return

				H = gaussian_filter(H, sigma=self.bw)
				if self.cut_below is None:
					Hmin = H[H > 0][0]
				else:
					Hmin = self.cut_below

				if self.clip_above is None:
					self.Hmax = H.max()
				else:
					self.Hmax = self.clip_above

				if self.scaling == 'ranking':
					from statsmodels.distributions.empirical_distribution import ECDF
					ecdf = ECDF(H.flatten())

				for ix in range(len(xgrid) - 2):
					for iy in range(len(ygrid) - 2):
						if H[iy, ix] > Hmin:
							rects_vertices.append((xgrid[ix], ygrid[iy], xgrid[ix + 1], ygrid[iy + 1]))
							if self.scaling == 'ranking':
								rects_colors.append(self.cmap.to_color(ecdf(H[iy, ix]) - ecdf(Hmin), 1 - ecdf(Hmin), 'lin'))
							else:
								rects_colors.append(self.cmap.to_color(H[iy, ix], self.Hmax, self.scaling))
			else:
				raise Exception('method not supported')

			self.painter.batch_rects(rects_vertices, rects_colors)
		except:
			print "No train tweets for this timeslot."

	def draw(self, proj, mouse_x, mouse_y, ui_manager):
		self.painter.batch_draw()
		if self.show_colorbar:
			ui_manager.add_colorbar(self.cmap, self.Hmax, self.scaling)
		self.communicator.add_infobar(ui_manager, "DE")

	def bbox(self):
		return BoundingBox.from_points(lons=self.data['lon'], lats=self.data['lat'])

	def on_key_release(self, key, modifiers):
		self.communicator.on_key_release(key, modifiers)

	def update_data(self, data):
		self.data = data
