'''
    TweeLoc is created by

    Pavlos Paraskevopoulos:
        George Mason University
        LIPADE, Paris Descartes University
        pparaske@gmu.edu

    Giovanni Pellegrini:
        University of Trento
        giovanni.pellegrini@unitn.it

    Themis Palpanas:
        Paris Descartes University
        themis@mi.parisdescartes.fr

	Copyright Paraskevopoulos, Pellegrini, Palpanas.


	This software cannot be used for commercial purposes.

	You may use the software as is, only for research purposes.
	You should acknowledge the authors by referencing the following two papers (the first describes the method and algorithms; the second describes the system and visualizations):

	Pavlos Paraskevopoulos, Themis Palpanas.
	Where has this Tweet Come From? Fast and Fine-Grained Geolocalisation of Non-Geotagged Tweets. Social Network Analysis and Mining (SNAM) Journal 6(1), pp 1-16, 2016.
	http://www.mi.parisdescartes.fr/~themisp/publications/snam16.pdf

	Pavlos Paraskevopoulos, Giovanni Pellegrini, Themis Palpanas. TweeLoc: A System for Geolocalizing Tweets at Fine-Grain. International Conference on Data Mining (ICDM), New Orleans, LA, USA, November 2017.
	http://www.mi.parisdescartes.fr/~themisp/publications/icdm17-tweeloc.pdf
'''

import numpy as np
from geoplotlib.layers import BaseLayer, HotspotManager
from geoplotlib.core import BatchPainter
import os


MASK = pow(2, 16) - 1


def index_to_id(r, c):
	index_r = (r & MASK) << 16
	index_c = c & MASK
	id = index_r | index_c
	return id


class HeatMap(BaseLayer):
	def __init__(self, keywords=None, counts=None, coordinates=None, communicator=None):
		self.keywords = keywords
		self.counts = counts
		self.coordinates = coordinates
		self.communicator = communicator
		self.hotspots = HotspotManager()
		self.painter = BatchPainter()
		self.max_v = -100000000
		self.min_v = 100000000
		self.gradient_test = {'r': [92, 82, 73, 64, 55, 46, 36, 27, 18, 9, 0],
							  'g': [202, 182, 163, 144, 125, 106, 87, 68, 49, 30, 11],
							  'b': [255, 246, 237, 228, 220, 211, 202, 194, 185, 176, 168]}

	def invalidate(self, proj):
		self.painter = BatchPainter()
		self.hotspots = HotspotManager()
		self.get_min_max()
		for name, sq_ids in self.coordinates.items():
			c_id = -1
			for c in sq_ids[0]:
				r_id = 0
				c_id += 1
				for r in sq_ids[1]:
					sq_id = index_to_id(r_id, c_id)
					r_id += 1
					lon = np.array([c[0], c[1]])
					lat = np.array([r[0], r[1]])
					x, y = proj.lonlat_to_screen(lon, lat)
					w = x.max() - x.min()
					h = y.max() - y.min()
					w = np.ceil(w / 2) * 2
					h = np.ceil(h / 2) * 2
					l = max(w, h)
					n_x = np.array([x.min(), x.min() + l, x.min() + l, x.min()])
					n_y = np.array([y.min() + l, y.min() + l, y.min(), y.min()])
					self.painter.set_color([0, 0, 0])
					self.painter.linestrip(n_x, n_y, closed=True)
					if sq_id in self.counts[name].keys():
						v = self.counts[name][sq_id]
					else:
						continue

					if v == 0:
						continue
					alpha = 100 + ((v * 100) / self.max_v)
					b = (abs(v) * 9 / self.max_v)
					col_c = [255, 255, 255, 255]
					try:
						if v > 0:
							rgb = [self.gradient_test['r'][int(b)],
								   self.gradient_test['g'][int(b)],
								   self.gradient_test['b'][int(b)]]
						col_c = rgb
						col_c.append(alpha)
					except:
						print "B= ", 9 - int(b)
					self.painter.set_color(col_c)
					self.painter.poly(n_x, n_y)
					try:
						self.hotspots.add_poly(n_x, n_y, self.keywords[name][sq_id])
						self.painter.set_color([0, 0, 0])
						self.painter.labels([x[0]], [y[0]], str(v), font_size=14, anchor_x='left',
										anchor_y='top')
					except:
						pass

	def draw(self, proj, mouse_x, mouse_y, ui_manager):
		self.painter.batch_draw()

	def on_key_release(self, key, modifiers):
		self.communicator.on_key_release(key, modifiers)

	def on_mouse_release(self, proj, mouse_x, mouse_y, modifiers):
		p_lat, p_lon = proj.screen_to_latlon(mouse_x, mouse_y)
		for name, sq_ids in self.coordinates.items():
			c_id = -1
			for c in sq_ids[0]:
				r_id = 0
				c_id += 1
				for r in sq_ids[1]:
					sq_id = index_to_id(r_id, c_id)
					r_id += 1
					if (c[0] <= p_lon < c[1]) and (r[1] <= p_lat < r[0]):
						self.communicator.on_mouse_release(name, sq_id, modifiers)

	def update_data(self, keywords, counts):
		self.keywords = keywords
		self.counts = counts

	def get_min_max(self):
		for name, sq_ids in self.coordinates.items():
			c_id = -1
			for c in sq_ids[0]:
				r_id = 0
				c_id += 1
				for r in sq_ids[1]:
					sq_id = index_to_id(r_id, c_id)
					r_id += 1
					if sq_id in self.counts[name].keys():
						v = self.counts[name][sq_id]
						if v > self.max_v:
							self.max_v = v
						elif v < self.min_v:
							self.min_v = v


class TrainHeatMap(HeatMap):
	def __init__(self, keywords=None, counts=None, coordinates=None, communicator=None):
		HeatMap.__init__(self, keywords, counts, coordinates, communicator)

	def draw(self, proj, mouse_x, mouse_y, ui_manager):
		self.painter.batch_draw()
		picked = self.hotspots.pick(mouse_x, mouse_y)
		if picked:
			ui_manager.tooltip(picked)
		self.communicator.add_infobar(ui_manager, "TR")


class PredictionHeatMap(HeatMap):
	def __init__(self, keywords=None, counts=None, coordinates=None, communicator=None):
		HeatMap.__init__(self, keywords, counts, coordinates, communicator)

	def draw(self, proj, mouse_x, mouse_y, ui_manager):
		self.painter.batch_draw()
		picked = self.hotspots.pick(mouse_x, mouse_y)
		if picked:
			ui_manager.tooltip(picked)
		self.communicator.add_infobar(ui_manager, "PR")


class DifferentialHeatMap(HeatMap):
	def __init__(self, keywords=None, counts=None, coordinates=None, communicator=None):
		HeatMap.__init__(self, keywords, counts, coordinates, communicator)
		self.gradient_pos = {'r': [0, 17, 35, 52, 70, 87, 105, 122, 140, 157, 175],
							 'g': [229, 231, 234, 236, 239, 241, 244, 246, 249, 251, 254],
							 'b': [38, 49, 60, 71, 82, 93, 104, 115, 126, 137, 149]}
		self.gradient_neg = {'r': [229, 231, 234, 236, 239, 241, 244, 246, 249, 251, 254],
							 'g': [2, 19, 37, 55, 72, 90, 108, 125, 143, 161, 179],
							 'b': [0, 14, 29, 44, 59, 74, 89, 104, 119, 134, 149]}

	def invalidate(self, proj):
		self.painter = BatchPainter()
		self.get_min_max()
		for name, sq_ids in self.coordinates.items():
			c_id = -1
			for c in sq_ids[0]:
				r_id = 0
				c_id += 1
				for r in sq_ids[1]:
					sq_id = index_to_id(r_id, c_id)
					r_id += 1
					lon = np.array([c[0], c[1]])
					lat = np.array([r[0], r[1]])
					x, y = proj.lonlat_to_screen(lon, lat)
					w = x.max() - x.min()
					h = y.max() - y.min()
					w = np.ceil(w / 2) * 2
					h = np.ceil(h / 2) * 2
					l = max(w, h)
					n_x = np.array([x.min(), x.min() + l, x.min() + l, x.min()])
					n_y = np.array([y.min() + l, y.min() + l, y.min(), y.min()])
					self.painter.set_color([0, 0, 0])
					self.painter.linestrip(n_x, n_y, closed=True)
					if sq_id in self.counts[name].keys():
						v = self.counts[name][sq_id]
					else:
						continue

					alpha = int(abs(100 + ((v * 100) / 200)))
					if self.max_v == 0:
						b = 0
					else:
						b = (abs(v) * 9 / self.max_v)
					try:
						if v < 0:
							# b = b- min_v
							rgb = [self.gradient_neg['r'][9 - int(b)],
								   self.gradient_neg['g'][9 - int(b)],
								   self.gradient_neg['b'][9 - int(b)]]
						# col_c = rgb
						elif v > 0:
							rgb = [self.gradient_pos['r'][9 - int(b)],
								   self.gradient_pos['g'][9 - int(b)],
								   self.gradient_pos['b'][9 - int(b)]]
						# col_c = rgb
						elif v == 0:
							rgb = [255, 255, 255]
						# col_c = rgb
						col_c = rgb
						if alpha != 0:
							col_c.append(150)
						else:
							col_c.append(alpha + 10)
					except:
						print "B= ", 9 - int(b)
					self.painter.set_color(col_c)
					self.painter.poly(n_x, n_y)
					self.painter.set_color([0, 0, 0])
					self.painter.labels([x[0]], [y[0]], str(v) + "%", font_size=14, anchor_x='left',
										anchor_y='top')

	def draw(self, proj, mouse_x, mouse_y, ui_manager):
		self.painter.batch_draw()
		picked = self.hotspots.pick(mouse_x, mouse_y)
		if picked:
			ui_manager.tooltip(picked)
		self.communicator.add_infobar(ui_manager, "DI")
