'''
    TweeLoc is created by

    Pavlos Paraskevopoulos:
        George Mason University
        LIPADE, Paris Descartes University
        pparaske@gmu.edu

    Giovanni Pellegrini:
        University of Trento
        giovanni.pellegrini@unitn.it

    Themis Palpanas:
        Paris Descartes University
        themis@mi.parisdescartes.fr

    Copyright Paraskevopoulos, Pellegrini, Palpanas.


    This software cannot be used for commercial purposes.

    You may use the software as is, only for research purposes.
    You should acknowledge the authors by referencing the following two papers (the first describes the method and algorithms; the second describes the system and visualizations):

    Pavlos Paraskevopoulos, Themis Palpanas.
    Where has this Tweet Come From? Fast and Fine-Grained Geolocalisation of Non-Geotagged Tweets. Social Network Analysis and Mining (SNAM) Journal 6(1), pp 1-16, 2016.
    http://www.mi.parisdescartes.fr/~themisp/publications/snam16.pdf

    Pavlos Paraskevopoulos, Giovanni Pellegrini, Themis Palpanas. TweeLoc: A System for Geolocalizing Tweets at Fine-Grain. International Conference on Data Mining (ICDM), New Orleans, LA, USA, November 2017.
    http://www.mi.parisdescartes.fr/~themisp/publications/icdm17-tweeloc.pdf
'''

import random
import math
from itertools import imap
import sys


class Prediction:
    def __init__(self, thresholds, percentage):
        self.predictions_l1 = [{} for t in thresholds]
        self.predictions_l2 = [{} for t in thresholds]
        self.similarities_ql = {}
        self.predictions_l1_position = None
        self.predictions_l2_position = None
        self.probabilities = [{} for t in thresholds]
        self.top_matches = [{} for t in thresholds]
        self.results = self._create_result_structure(percentage, thresholds)

    @staticmethod
    def _create_result_structure(percentage, thresh):
        results = {}
        for p in percentage:
            results[p] = {}
            for t in thresh:
                results[p].update({t: {'precision': []}})
                results[p][t].update({'recall': []})
                results[p][t].update({'f1': []})
        return results


# Create the structure to store the thresholds
def create_threshold_structure(n_key, ts_per_day):
    thrs = {}
    for i in n_key:
        thrs[i] = {}
        for ts_id in xrange(0, ts_per_day - 1, 2):
            thrs[i][ts_id] = [[0.0], [0.0]]
    return thrs


# Divide and return the dataset splitted into train and testing.
def split_dataset_train_test(tweets, train_percent, start_ts, window):
    train_set = {}
    test_set = {}
    total_test = 0
    total_train = 0
    for l1_id, city in tweets.items():
        temp_train_set = {}
        temp_test_set = {}
        for sq_id, item in city.items():
            tweets_sq = []
            for ts_id, tweet_list in item.items():
                if start_ts <= ts_id < start_ts + window:
                    [tweets_sq.append(x) for x in tweet_list]
                else:
                    continue
            n_tweets = len(tweets_sq)
            if n_tweets == 0:
                continue
            # Add a small quantity otherwise there are problem with conversion between float and int
            test_size = int(n_tweets * (1 - train_percent) + 0.0000001)
            if test_size >= 1:
                train_size = n_tweets - test_size
            else:
                train_size = n_tweets
            total_train += train_size
            total_test += test_size
            random.shuffle(tweets_sq)
            if test_size >= 1:
                temp_train_set[sq_id] = tweets_sq[:train_size]
                temp_test_set[sq_id] = tweets_sq[train_size:]
            else:
                temp_train_set[sq_id] = tweets_sq
        train_set[l1_id] = temp_train_set
        test_set[l1_id] = temp_test_set
    return train_set, test_set, total_test


def local_global_activities(tweets, window, start_ts, level):
    local_activities = {}
    global_activity = [0 for y in xrange(window)]
    if level == "L1":
        for c_id in tweets.keys():
            local_activities[c_id] = [0 for y in xrange(window)]
        current_ts = start_ts
        for i in xrange(window):
            sum_global = 0
            for c_id, city in tweets.items():
                n_tweets = 0
                for sq_id, sq_list in city.items():
                    if current_ts in sq_list.keys():
                        n_tweets += len(tweets[c_id][sq_id][current_ts])
                        sum_global += n_tweets
                local_activities[c_id][i] = n_tweets
            global_activity[i] = sum_global
            current_ts += 1
    if level == "L2":
        for sq_id in tweets.keys():
            local_activities[sq_id] = [0 for y in xrange(window)]
        current_ts = start_ts
        for i in xrange(window):
            sum_global = 0
            for sq_id, item in tweets.items():
                if current_ts in item.keys():
                    n_tweets = len(tweets[sq_id][current_ts])
                    sum_global += n_tweets
                    local_activities[sq_id][i] = n_tweets
            global_activity[i] = sum_global
            current_ts += 1
    return local_activities, global_activity


def correlations(local_activities, global_activity, window):
    correlation_values = {}
    counter_lambda_g = 0
    half_window = window / 2
    z = [v for v in range(1, window + 1)]
    for i in xrange(half_window):
        lambda_g = _linear_regression_lambda(global_activity[i:i + half_window], z[i:i + half_window])
        if lambda_g > 0:
            counter_lambda_g += 1

    if counter_lambda_g > 0:
        for id, activity in local_activities.items():
            counter_lambda_l = 0
            for i in xrange(half_window):
                lambda_l = _linear_regression_lambda(activity[i:i + half_window], z[i:i + half_window])
                if lambda_l > 0:
                    counter_lambda_l += 1
            p_correlation = _pearsonr(activity, global_activity) + 1
            if counter_lambda_l > 0:
                correlation_values[id] = {'trend': True, 'pearson': p_correlation}
            else:
                correlation_values[id] = {'trend': False, 'pearson': p_correlation}
    else:
        for id in local_activities.keys():
            correlation_values[id] = {'trend': False, 'pearson': 0}
    return correlation_values


def _mean(x):
    return sum(x) / len(x)


def _linear_regression_lambda(x, y):
    sum_xy = 0
    sum_sq_v_x = 0
    sum_sq_v_y = 0

    m_X = _mean(x)
    m_Y = _mean(y)
    for (x, y) in zip(x, y):
        var_x = x - m_X
        var_y = y - m_Y
        sum_xy += var_x * var_y
        sum_sq_v_x += pow(var_x, 2)
        sum_sq_v_y += pow(var_y, 2)
    if math.sqrt(sum_sq_v_x * sum_sq_v_y) == 0:
        return 0
    return sum_xy / math.sqrt(sum_sq_v_x * sum_sq_v_y)


def _pearsonr(x, y):
    # Assume len(x) == len(y)
    n = len(x)
    sum_x = float(sum(x))
    sum_y = float(sum(y))
    sum_x_sq = sum(map(lambda x: pow(x, 2), x))
    sum_y_sq = sum(map(lambda x: pow(x, 2), y))
    psum = sum(imap(lambda x, y: x * y, x, y))
    num = psum - (sum_x * sum_y / n)
    den = pow((sum_x_sq - pow(sum_x, 2) / n) * (sum_y_sq - pow(sum_y, 2) / n), 0.5)
    if den == 0: return 0
    return num / den


def create_tfidf_documents(t_set, level):
    if level == 'L1':
        words_in_documents = {}
        documents = {}
        for l1_id, l2_list in t_set.items():
            document = {}
            total_occurrences = 0
            for l2_id, tweet_list in l2_list.items():
                for tweet in tweet_list:
                    for word in tweet.text.split(' '):
                        if word not in words_in_documents.keys():
                            words_in_documents[word] = []
                        if l1_id not in words_in_documents[word]:
                            words_in_documents[word].append(l1_id)
                        if word not in document.keys():
                            document[word] = 1.0
                        else:
                            document[word] += 1.0
                        total_occurrences += 1
            document_tf = _calculate_tf(document, total_occurrences)
            documents[l1_id] = document_tf
        n_docs = len(documents.keys())
        words_idf = _calculate_idf(words_in_documents, n_docs)
        documents_tfidf = {}
        for id, document in documents.items():
            document_tfidf = _calculate_tfidf(document, words_idf)
            documents_tfidf[id] = document_tfidf

    elif level == 'L2':
        words_in_documents = {}
        documents = {}
        for id, tweet_list in t_set.items():
            document = {}
            total_occurrences = 0
            for tweet in tweet_list:
                for word in tweet.text.split(' '):
                    if word not in words_in_documents.keys():
                        words_in_documents[word] = []
                    if id not in words_in_documents[word]:
                        words_in_documents[word].append(id)
                    if word not in document.keys():
                        document[word] = 1.0
                    else:
                        document[word] += 1.0
                    total_occurrences += 1
            document_tf = _calculate_tf(document, total_occurrences)
            documents[id] = document_tf

        n_docs = len(documents.keys())
        words_idf = _calculate_idf(words_in_documents, n_docs)
        documents_tfidf = {}
        for id, document in documents.items():
            document_tfidf = _calculate_tfidf(document, words_idf)
            documents_tfidf[id] = document_tfidf

    return documents_tfidf, words_idf


def _calculate_tf(document, total):
    for word, occurrence in document.items():
        tf = float(occurrence) / float(total)
        document[word] = tf
    return document


def _calculate_tfidf(document, words_idf):
    for word, tf in document.items():
        document[word] *= words_idf[word]
    return document


def _calculate_idf(words, n_docs):
    words_idf = {}
    for word, docs in words.items():
        v = math.log(float(n_docs) / float(len(docs)), math.e)
        '''if v == 0:
            v = 0.0000000000001'''
        words_idf.update({word: v})
    return words_idf


# Delete keywords with lowest tfidf value, according to a certain percentage
def prune_documents(documents, percentage):
    pruned_docs = {}
    percentage = 100 - percentage
    for id, document in documents.iteritems():
        keywords_to_prune = int(round((percentage * len(document)) / 100))
        sorted_keys = sorted(document, key=document.__getitem__)
        if keywords_to_prune == 0:
            if not sorted_keys:
                pruned_docs[id] = document
                continue
        else:
            for i in range(keywords_to_prune):
                del document[sorted_keys[i]]
        pruned_docs[id] = document
    return pruned_docs


def print_progress(iteration, total, prefix='', suffix='', decimals=1, barLength=100):
    formatStr = "{0:." + str(decimals) + "f}"
    percents = formatStr.format(100 * (iteration / float(total)))
    filledLength = int(round(barLength * iteration / float(total)))
    bar = '|' * filledLength + '-' * (barLength - filledLength)
    sys.stdout.write('\r%s |%s| %s%s %s  ' % (prefix, bar, percents, '%', suffix))
    sys.stdout.flush()


def predict_tweets_l1_loc(conf, test, documents, idf, thrs, percentage, start_ts, correlations):
    similarities = {}
    for t in conf.thresholds:
        similarities[t] = {}
    # similarities = [{t:{}} for t in conf.thresholds]
    # similarities_no_thresh = {}
    for l2_list in test.values():
        for tweet_list in l2_list.values():
            for tweet in tweet_list:
                q_tfidf = create_query_tfidf(tweet, idf)
                for doc_id, document in documents.iteritems():
                    similarity = relation(q_tfidf, document)
                    if correlations:
                        if correlations[doc_id]['trend']:
                            cor_v = correlations[doc_id]['pearson']
                            similarity *= cor_v
                            thresh = sum(thrs[percentage][start_ts % conf.n_ts_per_day][0]) / len(
                                thrs[percentage][start_ts % conf.n_ts_per_day][0])
                            i = 0
                            for factor in conf.thresholds:
                                add_thrs = thresh * factor / 100.0
                                new_thresh = thresh + add_thrs
                                if similarity > new_thresh:
                                    if tweet.id not in similarities[factor].keys():
                                        similarities[factor].update({tweet.id: {doc_id: similarity}})
                                    else:
                                        similarities[factor][tweet.id].update({doc_id: similarity})
                                i += 1
                            '''if no_thresh and similarity > 0:
                                if not similarities_no_thresh.has_key(tweet.id):
                                    similarities_no_thresh.update({tweet.id:{doc_id:similarity}})
                                else:
                                    similarities_no_thresh[tweet.id].update({doc_id:similarity})'''

                    else:
                        thresh = sum(thrs[percentage][start_ts % conf.n_ts_per_day][0]) / len(
                            thrs[percentage][start_ts % conf.n_ts_per_day][0])
                        i = 0
                        for factor in conf.thresholds:
                            add_thrs = thresh * factor / 100.0
                            new_thresh = thresh + add_thrs
                            # thrs_index = str(factor)
                            if similarity > new_thresh:
                                if tweet.id not in similarities[factor].keys():
                                    similarities[factor].update({tweet.id: {doc_id: similarity}})
                                else:
                                    similarities[factor][tweet.id].update({doc_id: similarity})
                            i += 1
                        '''if no_thresh and similarity > 0:
                            if not similarities_no_thresh.has_key(tweet.id):
                                similarities_no_thresh.update({tweet.id: {doc_id: similarity}})
                            else:
                                similarities_no_thresh[tweet.id].update({doc_id: similarity})'''
    return similarities


def predict_tweets_l2_loc(conf, test, documents, idf, thrs, percentage, start_ts, correlations, loc_id, similarities):
    for l2_list in test.values():
        for tweet_list in l2_list.values():
            for tweet in tweet_list:
                q_tfidf = create_query_tfidf(tweet, idf)

                for doc_id, document in documents.iteritems():
                    similarity = relation(q_tfidf, document)
                    if correlations:
                        if correlations[loc_id][doc_id]['trend']:
                            cor_v = correlations[loc_id][doc_id]['pearson']
                            similarity *= cor_v
                            thresh = sum(thrs[percentage][start_ts % conf.n_ts_per_day][1]) / len(
                                thrs[percentage][start_ts % conf.n_ts_per_day][1])
                            i = 0
                            for factor in conf.thresholds:
                                add_thrs = thresh * factor / 100.0
                                new_thresh = thresh + add_thrs
                                # factor = str(factor)
                                if similarity > new_thresh:
                                    if tweet.id not in similarities[factor].keys():
                                        similarities[factor].update({tweet.id: {loc_id: {doc_id: similarity}}})
                                    else:
                                        if loc_id not in similarities[factor][tweet.id].keys():
                                            similarities[factor][tweet.id].update({loc_id: {doc_id: similarity}})
                                        else:
                                            similarities[factor][tweet.id][loc_id].update({doc_id: similarity})
                                i += 1
                            '''if no_thresh and similarity > 0:
                                if not similarities_no_thresh.has_key(tweet.id):
                                    similarities_no_thresh.update({tweet.id: {loc_id: {doc_id:similarity}}})
                                else:
                                    if not similarities_no_thresh[tweet.id].has_key(loc_id):
                                        similarities_no_thresh[tweet.id].update({loc_id: {doc_id:similarity}})
                                    else:
                                        similarities_no_thresh[tweet.id][loc_id].update({doc_id:similarity})'''

                    else:
                        thresh = sum(thrs[percentage][start_ts % conf.n_ts_per_day][1]) / len(
                            thrs[percentage][start_ts % conf.n_ts_per_day][1])
                        i = 0
                        for factor in conf.thresholds:
                            add_thrs = thresh * factor / 100.0
                            new_thresh = thresh + add_thrs
                            # thrs_index = str(factor)
                            if similarity > new_thresh:
                                if tweet.id not in similarities[factor].keys():
                                    similarities[factor].update({tweet.id: {loc_id: {doc_id: similarity}}})
                                else:
                                    if loc_id not in similarities[factor][tweet.id].keys():
                                        similarities[factor][tweet.id].update({loc_id: {doc_id: similarity}})
                                    else:
                                        similarities[factor][tweet.id][loc_id].update({doc_id: similarity})
                            i += 1
                        '''if no_thresh and similarity > 0:
                            if not similarities_no_thresh.has_key(tweet.id):
                                similarities_no_thresh.update({tweet.id: {loc_id: {doc_id:similarity}}})
                            else:
                                if not similarities_no_thresh[tweet.id].has_key(loc_id):
                                    similarities_no_thresh[tweet.id].update({loc_id: {doc_id:similarity}})
                                else:
                                    similarities_no_thresh[tweet.id][loc_id].update({doc_id:similarity})'''
    return similarities


def create_query_tfidf(tweet, idf):
    q_tfidf = {}
    total_words = 0
    for word in tweet.text.split(' '):
        if len(word) > 2:
            if word in q_tfidf.keys():
                q_tfidf[word] += 1.0
            else:
                q_tfidf[word] = 1.0
            total_words += 1
    for word in q_tfidf.keys():
        if word in idf.keys():
            q_tfidf[word] = (q_tfidf[word] / float(total_words)) * idf[word]
        else:
            q_tfidf[word] = 0
    return q_tfidf


def relation(query, document):
    numerator = 0.0
    for word in query.keys():
        if word in document.keys():
            numerator += query[word] * document[word]
    q_mag = magnitude(query)
    d_mag = magnitude(document)
    denominator = q_mag * d_mag
    if denominator > 0:
        return numerator / denominator
    else:
        return 0.0


def magnitude(vector):
    total = 0.0
    for word, count in vector.iteritems():
        total += count ** 2
    return math.sqrt(total)


def merge_probabilities(l1_prob, l2_prob):
    final_probabilities = {}
    for tweet_id, prob in l2_prob.iteritems():
        if not prob:
            final_probabilities[tweet_id] = {}
        else:
            for l1_id in prob.keys():
                if tweet_id in l1_prob.keys():
                    if l1_id in l1_prob[tweet_id].keys():
                        for l2_id in prob[l1_id].keys():
                            final_probability = l1_prob[tweet_id][l1_id] * l2_prob[tweet_id][l1_id][l2_id]
                            if tweet_id not in final_probabilities.keys():
                                final_probabilities.update({tweet_id: {}})
                            if l1_id not in final_probabilities[tweet_id].keys():
                                final_probabilities[tweet_id].update({l1_id: {}})
                            final_probabilities[tweet_id][l1_id].update({l2_id: final_probability})
                    else:
                        continue
                else:
                    continue
    return final_probabilities


def extract_most_probable_place(probabilities):
    tweet_and_place = {}

    for tweet_id, l1_list in probabilities.iteritems():
        top = 0.0
        place = ''
        for l1_id, l2_list in l1_list.iteritems():
            for l2_id, prob in l2_list.iteritems():
                if top < prob:
                    top = prob
                    place = str(l1_id) + ' ' + str(l2_id) + ' ' + str(prob)
        if top > 0.0000000000001 and place != '':
            tweet_and_place[tweet_id] = place

    return tweet_and_place


def update_thresholds(conf, thresholds, test, prediction, prediction_l1, prediction_l2, percentage, timeslot):
    ok = 0
    avg_l1 = 0.0
    avg_l2 = 0.0
    for l1_id, l2_list in test.iteritems():
        for l2_id, tweet_list in l2_list.iteritems():
            for tweet in tweet_list:
                if tweet.id in prediction.keys():
                    info = prediction[tweet.id].split(" ")
                    if info[0] != '':
                        if l1_id == info[0] and l2_id == int(info[1]):
                            ok += 1
                            avg_l1 += prediction_l1[tweet.id][l1_id]
                            avg_l2 += prediction_l2[tweet.id][l1_id][l2_id]
    if ok != 0:
        avg_l1 /= ok
        avg_l2 /= ok
        thresholds[percentage][timeslot % conf.n_ts_per_day][0].append(avg_l1)
        thresholds[percentage][timeslot % conf.n_ts_per_day][1].append(avg_l2)
    return thresholds


def confront_prediction_actual_test(test, prediction, n_test):
    ok = 0
    not_ok = 0
    for l1_id, l2_list in test.iteritems():
        for l2_id, tweet_list in l2_list.iteritems():
            for tweet in tweet_list:
                if tweet.id in prediction.keys():
                    info = prediction[tweet.id].split(" ")
                    if info[0] != '':
                        if l1_id == info[0] and l2_id == int(info[1]):
                            ok += 1
                        else:
                            not_ok += 1
                    else:
                        not_ok += 1

    if ok + not_ok == 0:
        return -1, -1, -1
    else:
        precision = round((float(ok) / float(ok + not_ok)) * 100, 2)
        recall = round((float(ok) / float(n_test)) * 100, 2)
        if precision + recall == 0:
            f1 = 0
        else:
            f1 = round(2 * (precision * recall) / (precision + recall), 2)
        return precision, recall, f1


def get_tweets_to_predict(tweets, start):
    tweets_ts = []
    for id in tweets.keys():
        if start <= id < start + 2:
            for tweet in tweets[id]:
                tweets_ts.append(tweet)
    return tweets_ts


def predict_tweets_l1_no_position(tweets, documents, idf, thrs, percentage, start_ts, correlations, ts_per_day):
    similarities = {}
    for tweet in tweets:
        q_tfidf = create_query_tfidf(tweet, idf)
        for doc_id, document in documents.iteritems():
            similarity = relation(q_tfidf, document)
            if correlations:
                if correlations[doc_id]['trend']:
                    cor_v = correlations[doc_id]['pearson']
                    similarity *= cor_v
                    thresh = sum(thrs[percentage][start_ts % ts_per_day][0]) / len(
                        thrs[percentage][start_ts % ts_per_day][0])
                    add_thrs = thresh * 10 / 100.0
                    new_thresh = thresh + add_thrs
                    if similarity > new_thresh:
                        if not similarities.has_key(tweet.id):
                            similarities.update({tweet.id: {doc_id: similarity}})
                        else:
                            similarities[tweet.id].update({doc_id: similarity})
            else:
                thresh = sum(thrs[percentage][start_ts % ts_per_day][0]) / len(
                    thrs[percentage][start_ts % ts_per_day][0])
                add_thrs = thresh * 10 / 100.0
                new_thresh = thresh + add_thrs
                if similarity > new_thresh:
                    if not similarities.has_key(tweet.id):
                        similarities.update({tweet.id: {doc_id: similarity}})
                    else:
                        similarities[tweet.id].update({doc_id: similarity})
    return similarities


def predict_tweets_l2_no_position(tweets, documents, idf, thrs, percentage, start_ts, correlations, ts_per_day,
                                  similarities, loc_id):
    similarities = {}
    for tweet in tweets:
        q_tfidf = create_query_tfidf(tweet, idf)
        for doc_id, document in documents.iteritems():
            similarity = relation(q_tfidf, document)
            if correlations:
                if correlations[loc_id][doc_id]['trend'] == True:
                    cor_v = correlations[loc_id][doc_id]['pearson']
                    similarity *= cor_v
                    thresh = sum(thrs[percentage][start_ts % ts_per_day][0]) / len(
                        thrs[percentage][start_ts % ts_per_day][0])
                    add_thrs = thresh * 10 / 100.0
                    new_thresh = thresh + add_thrs
                    if similarity > new_thresh:
                        if tweet.id not in similarities.keys():
                            similarities.update({tweet.id: {loc_id: {doc_id: similarity}}})
                        else:
                            if loc_id not in similarities[tweet.id].keys():
                                similarities[tweet.id].update({loc_id: {doc_id: similarity}})
                            else:
                                similarities[tweet.id][loc_id].update({doc_id: similarity})
            else:
                thresh = sum(thrs[percentage][start_ts % ts_per_day][0]) / len(
                    thrs[percentage][start_ts % ts_per_day][0])
                add_thrs = thresh * 10 / 100.0
                new_thresh = thresh + add_thrs
                if similarity > new_thresh:
                    if tweet.id not in similarities.keys():
                        similarities.update({tweet.id: {loc_id: {doc_id: similarity}}})
                    else:
                        if loc_id not in similarities[tweet.id].keys():
                            similarities[tweet.id].update({loc_id: {doc_id: similarity}})
                        else:
                            similarities[tweet.id][loc_id].update({doc_id: similarity})
    return similarities
