'''
    TweeLoc is created by

    Pavlos Paraskevopoulos:
        George Mason University
        LIPADE, Paris Descartes University
        pparaske@gmu.edu

    Giovanni Pellegrini:
        University of Trento
        giovanni.pellegrini@unitn.it

    Themis Palpanas:
        Paris Descartes University
        themis@mi.parisdescartes.fr

    Copyright Paraskevopoulos, Pellegrini, Palpanas.


    This software cannot be used for commercial purposes.

    You may use the software as is, only for research purposes.
    You should acknowledge the authors by referencing the following two papers (the first describes the method and algorithms; the second describes the system and visualizations):

    Pavlos Paraskevopoulos, Themis Palpanas.
    Where has this Tweet Come From? Fast and Fine-Grained Geolocalisation of Non-Geotagged Tweets. Social Network Analysis and Mining (SNAM) Journal 6(1), pp 1-16, 2016.
    http://www.mi.parisdescartes.fr/~themisp/publications/snam16.pdf

    Pavlos Paraskevopoulos, Giovanni Pellegrini, Themis Palpanas. TweeLoc: A System for Geolocalizing Tweets at Fine-Grain. International Conference on Data Mining (ICDM), New Orleans, LA, USA, November 2017.
    http://www.mi.parisdescartes.fr/~themisp/publications/icdm17-tweeloc.pdf
'''

import os
import sys
import glob
import time
import json
import locale
import re
import geopy
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import random
import operator
import pickle
from geopy.distance import VincentyDistance
from datetime import datetime, timedelta


# Tweet object class
class Tweet:
    def __init__(self, id, lang, time, text, position):
        self.id = id
        self.time = time
        self.text = text
        self.position = position
        self.language = lang

    def toJson(self):
        return "{\"created_at\":\"" + self.time.__str__() + \
               "\",\"id\":" + self.id + \
               ",\"id_str\":\"" + self.id + \
               "\",\"lang\":\"" + self.language + \
               "\",\"text\":\"" + self.text + \
               "\",\"coordinates\":[" + str(self.position['lon']) + "," + str(self.position['lat']) + "]}"


class City:
    def __init__(self, name, lat, lon, x_side, y_side):
        self.name = name
        self.lat = lat
        self.lon = lon
        self.x_side = x_side
        self.y_side = y_side


class Conf:
    def __init__(self, folder, start_dt, stop_dt, locations, l2_side, window, timeslot_size, n_runs, train,
                 lang_filters, stop_words, thresholds, keywords_percentage, l1_corr, l2_corr, thresholds_file, stream,
                 consumer_key, consumer_secret, access_token_key, access_token_secret, bounding_box):
        self.folder = folder
        self.start_dt = start_dt
        self.stop_dt = stop_dt
        self.locations = locations
        self.l2_side = l2_side
        self.window = window
        self.timeslot_size = timeslot_size
        self.n_runs = n_runs
        self.train = train
        self.lang_filters = lang_filters
        self.stop_words = stop_words
        self.thresholds = thresholds
        self.keywords_percentage = keywords_percentage
        self.l1_correlation = l1_corr
        self.l2_correlation = l2_corr
        self.mask = pow(2, 16) - 1
        data_path = "/Run_" + time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(time.time()))
        dir = os.path.dirname(__file__)
        os.mkdir(dir + data_path)
        self.dir_path = dir + data_path
        self.thresholds_structure = thresholds_file
        self.stream = stream
        self.n_ts_per_day = 1440 / self.timeslot_size
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.access_token_key = access_token_key
        self.access_token_secret = access_token_secret
        self.bounding_box = bounding_box

    def print_Conf(self):
        print "========== Settings of the program =============\n"
        print "========== Basic settings ======================"
        print "input_folder:\t\t" + str(self.folder)
        print "start_datetime:\t\t" + str(self.start_dt)
        print "stop_datetime:\t\t" + str(self.stop_dt)
        print "input_l1:"
        for location in self.locations:
            print "________________________________________________________"
            print "\t\t{0}".format(str(location.name))
            print "\t\t\tcoordinates: {0}, {1}".format(str(location.lat), str(location.lon))
            print "\t\t\tsize: {0} x, {1} y (meters)".format(str(location.x_side), str(location.y_side))
        print "\n========== Algorithm settings ================="
        print "l2_square_side:\t\t" + str(self.l2_side)
        print "window_size:\t\t" + str(self.window)
        print "timeslot_size:\t\t" + str(self.timeslot_size)
        print "n_runs:\t\t" + str(self.n_runs)
        print "train_percentage:\t\t" + str(self.train)
        print "languages_filter:\t\t" + str(self.lang_filters)
        print "stop_words:\t\t" + str(self.stop_words)
        print "thresholds_list:\t\t" + str(self.thresholds)
        print "keywords_percentage:\t\t" + str(self.keywords_percentage)

        print "\n================= Test type ==================="
        print "l1_correlations:\t\t" + str(self.l1_correlation)
        print "l2_correlations:\t\t" + str(self.l2_correlation)

        print "\n============= Thresholds file ================="
        print "thresholds:\t\t" + str(self.thresholds_structure)

        print "\n============= Visualization ==================="
        print "stream:\t\t" + str(self.stream)

        print "\n============= Twitter credentials ==================="
        print "consumer_key:\t\t" + str(self.consumer_key)
        print "consumer_secret:\t\t" + str(self.consumer_secret)
        print "access_token_key:\t\t" + str(self.access_token_key)
        print "access_token_secret:\t\t" + str(self.access_token_secret)

        print "\n============= Bounding Box ==================="
        print "bounding_box:\t\t" + str(self.bounding_box)

        print "\nResults and data will be saved in folder {0}".format(self.dir_path)

        print "\n"


def check_parameters_value():
    # Open file settings.txt and read all the values of the parameters
    consumer_key = ''
    consumer_secret = ''
    access_token_key = ''
    access_token_secret = ''
    with open("./settings.txt", "r") as settings_file:
        for line in settings_file:
            if line.startswith('#') or line.startswith('\n'):
                continue
            params = line[:-1].split('=')
            if params[0] == 'input_folder':
                input_folder = params[1]
            elif params[0] == 'start_datetime':
                datetime_1 = params[1]
            elif params[0] == 'stop_datetime':
                datetime_2 = params[1]
            elif params[0] == 'input_l1':
                l1_locations = params[1]
            elif params[0] == 'l2_square_side':
                square_side = params[1]
            elif params[0] == 'window_size':
                window_size = params[1]
            elif params[0] == 'timeslot_size':
                timeslot_size = params[1]
            elif params[0] == 'n_runs':
                n_runs = params[1]
            elif params[0] == 'train_percentage':
                train_percent = params[1]
            elif params[0] == 'languages_filter':
                lang_filter = params[1]
            elif params[0] == 'stop_words':
                stop_words = params[1].split(',')
            elif params[0] == 'thresholds_list':
                thresholds_str = params[1].split(',')
                thresholds_str.append('-110')
            elif params[0] == 'keywords_percentage':
                keywords_str = params[1].split(',')
                keywords_percentage = [int(k) for k in keywords_str]
            elif params[0] == 'l1_correlation':
                l1_correlation = params[1]
            elif params[0] == 'l2_correlation':
                l2_correlation = params[1]
            elif params[0] == 'thresholds_file':
                thresholds_load = params[1]
            elif params[0] == 'stream':
                stream = params[1]
            elif params[0] == 'consumer_key':
                consumer_key = params[1]
            elif params[0] == 'consumer_secret':
                consumer_secret = params[1]
            elif params[0] == 'access_token_key':
                access_token_key = params[1]
            elif params[0] == 'access_token_secret':
                access_token_secret = params[1]
            elif params[0] == 'bounding_box':
                bounding_box = params[1]

    # Check input folder to exists, not to be empty and files to be json
    dir = os.path.dirname(__file__)
    folder_path = os.path.join(dir, input_folder)
    if not os.path.isdir(folder_path):
        print "Invalid folder path: {0}.".format(folder_path)
        sys.exit(1)
    else:
        if os.listdir(folder_path):
            for tweet_file in glob.glob(folder_path + "*"):
                if tweet_file[-4:] != 'json':
                    raise Exception("{0} is not a json file".format(tweet_file))

    # Check start and stop date
    try:
        start_datetime = datetime.strptime(datetime_1, "%d-%b-%Y %H:%M:%S")
        stop_datetime = datetime.strptime(datetime_2, "%d-%b-%Y %H:%M:%S")
        if start_datetime >= stop_datetime:
            raise Exception()
    except Exception as e:
        raise Exception("Invalid start/stop datetime")

    # Check level 1 locations
    if not l1_locations:
        raise Exception("Invalid level 1 locations.")
    try:
        locations = l1_locations.split('|')
        locations_info = []
        for location in locations:
            name_info = location.split(':')
            params = name_info[1].split(',')
            locations_info.append(
                City(name_info[0], float(params[0]), float(params[1]), int(params[2]), int(params[3])))
    except:
        raise Exception(
            "Unable to parse locations. Values must be comma (,) separated, locations must be pipe (|) separated")

    # Check square side
    try:
        l2_side = int(square_side)
    except:
        raise Exception("Unable to convert square side")

    # Check window size
    try:
        window = int(window_size)
    except:
        raise Exception("Unable to convert window size")

    # Check timeslot size
    try:
        timeslot = int(timeslot_size)
    except:
        raise Exception("Unable to convert timeslot size")

    # Check n_runs
    try:
        runs = int(n_runs)
    except:
        raise Exception("Unable retrieve the number of runs")

    # Check train percentage
    try:
        train = float(train_percent)
    except:
        raise Exception("Unable to convert train value")
    if train < 0.1 or train > 0.99:
        raise Exception("Train value must be in range (0.1,0.99)")

    # Check language filter
    langs = lang_filter.split('|')
    filters = []
    for lang in langs:
        if not 2 < len(lang) < 4 and lang != 'none':
            print "Invalid language. Language name must be less than 3 characters."
            sys.exit(1)
        filters.append(lang)

    # Check if stopwords is none
    if stop_words == 'none' or stop_words == '':
        stop_words = None

    # Check thresholds
    try:
        thresholds_list = [int(k) for k in thresholds_str]
    except:
        raise Exception("Unable to convert thresholds list")

    # Check l1 and l2 correlation
    if l1_correlation.lower() == 'true':
        l1_corr = True
    elif l1_correlation.lower() == 'false':
        l1_corr = False
    else:
        raise Exception("l1 correlation is a boolean, true or false")

    if l2_correlation.lower() == 'true':
        l2_corr = True
    elif l2_correlation.lower() == 'false':
        l2_corr = False
    else:
        raise Exception("l1 correlation is a boolean, true or false")

    n_ts_per_day = 1440 / int(timeslot_size)
    if thresholds_load == 'none':
        thresholds_structure = None
    else:
        if not os.path.isfile(thresholds_load):
            raise Exception("Invalid path for thresholds file")
        else:
            fin = open(thresholds_load, 'rb')
            thresholds_structure = pickle.load(fin)
            fin.close()
            if sorted(thresholds_structure.keys()) != sorted(keywords_percentage):
                raise Exception(
                    "Invalid thresholds structure. Settings and structure in the file have different number of keywords percentage")
            elif len(thresholds_structure[max(keywords_percentage)].keys()) != n_ts_per_day / 2:
                raise Exception(
                    "Invalid thresholds structure. Settings and structure in the file have different number of time slots per day")

    # Check stream option
    if stream.lower() == 'true':
        stream_vis = True
    elif stream.lower() == 'false':
        stream_vis = False
    else:
        raise Exception("Stream is a boolean, true or false")

    bb = []
    for t in bounding_box.strip().split(','):
        bb.append(float(t))

    return Conf(folder_path, start_datetime, stop_datetime, locations_info, l2_side, window, timeslot, runs, train,
                filters, stop_words, thresholds_list, keywords_percentage, l1_corr, l2_corr, thresholds_structure,
                stream_vis, consumer_key, consumer_secret, access_token_key, access_token_secret, bb)


def create_folder_structure(conf):
    os.mkdir(conf.dir_path + "/Visualization_data")
    os.mkdir(conf.dir_path + "/Visualization_data/Squares_data")
    for location in conf.locations:
        os.mkdir(conf.dir_path + "/Visualization_data/Squares_data/" + location.name)
        os.mkdir(conf.dir_path + "/Visualization_data/Squares_data/" + location.name + "/Time_series")
    os.mkdir(conf.dir_path + "/Visualization_data/Points_data")
    os.mkdir(conf.dir_path + "/Visualization_data/Points_data/Train")
    os.mkdir(conf.dir_path + "/Visualization_data/Points_data/Test")

    os.mkdir(conf.dir_path + "/Predictions")

    os.mkdir(conf.dir_path + "/Results")

    os.mkdir(conf.dir_path + "/Plots")

    with open(conf.dir_path + "/Visualization_data/Run_info.txt", "w") as info_file:
        info_file.write("minutes_per_window:" + str(conf.timeslot_size * conf.window))


def create_timeslot_folder(conf, locations, date_time):
    for loc_id in locations:
        os.mkdir(conf.dir_path + "/Visualization_data/Squares_data/" + str(
            loc_id) + "/" + date_time.__str__())


def _decode_list(data):
    rv = []
    for item in data:
        if isinstance(item, unicode):
            item = item.encode('utf-8')
        elif isinstance(item, list):
            item = _decode_list(item)
        elif isinstance(item, dict):
            item = _decode_dict(item)
        rv.append(item)
    return rv


def _decode_dict(data):
    rv = {}
    for key, value in data.iteritems():
        if isinstance(key, unicode):
            key = key.encode('utf-8')
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        elif isinstance(value, list):
            value = _decode_list(value)
        elif isinstance(value, dict):
            value = _decode_dict(value)
        rv[key] = value
    return rv


# Takes in input the string of attribute "created_at" and returns a dict of the date elements
def parse_tweet_time(tweet_time):
    # extract the parts of the time from the string attribute "created_at" of the tweet
    locale.setlocale(locale.LC_TIME, 'en_US.utf8')
    new_tweet_time = datetime.strptime(tweet_time, '%a %b %d %H:%M:%S +0000 %Y')

    return new_tweet_time


def parse_tweet_position(tweet_position):
    return {'lon': tweet_position[0], 'lat': tweet_position[1]}


def parse_tweet_text(tweet_text):
    # text = tweet_text.encode('ascii', 'ignore')
    text_no_links = re.sub('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+', '',
                           tweet_text)
    parsed_text = re.sub(r'[^\w]', ' ', text_no_links)
    parsed_text = parsed_text.lower()
    return parsed_text.lower()


# Takes in input the folder of the tweets, returns the dict containing all the tweets
def retrieve_tweets_from_folder(conf):
    tweet_folder_path = conf.folder + '*'  # '*' is added to iterate files in the folder with glob
    # Counter for tweets with all informations and those without
    count_good_tweets = 0
    count_bad_tweets = 0
    count_pred_tweets = 0
    if not os.listdir(conf.folder):
        return
    # All files in the folder are read; from each file, extract the tweets and store
    # them into tweets dicitonary
    for tweet_file_name in glob.glob(tweet_folder_path):
        print("Reading file {0}".format(tweet_file_name))
        with open(tweet_file_name, 'r') as tweets_file:
            for tweet in tweets_file:
                if tweet == '\n':
                    continue
                else:
                    try:
                        tweet_json = json.loads(tweet, object_hook=_decode_dict)
                        id_str = tweet_json['id_str']
                        lang = tweet_json['lang']
                        time = parse_tweet_time(tweet_json['created_at'])  # time is a datetime type
                        text = parse_tweet_text(tweet_json['text'])  # text is parsed to keep only words
                        final_text = ""
                        for word in text.split(" "):
                            if len(word) > 3 and word not in conf.stop_words:
                                final_text += word + " "
                        try:
                            position = parse_tweet_position(
                                tweet_json['coordinates']['coordinates'])  # position is a dict composed of lat and lon
                            count_good_tweets += 1
                        except:
                            position = None
                            count_pred_tweets += 1
                        yield Tweet(id_str, lang, time, final_text, position)
                    except:
                        count_bad_tweets += 1
                        continue


# Extract the tweet done from start_time to stop_time
def extract_tweets_start_time_stop_time(tweets, start_time, stop_time):
    new_tweets = {}
    pred_tweets = {}
    for tweet in tweets:
        if start_time <= tweet.time <= stop_time:
            if tweet.position:
                new_tweets[tweet.id] = tweet
            else:
                pred_tweets[tweet.id] = tweet

    return new_tweets, pred_tweets


def extract_tweets_not_in_locations(tweets, tweets_cities):
    tweets_not_location = {}
    for sq_list in tweets_cities.values():
        for ts_list in sq_list.values():
            for tweet_list in ts_list.values():
                for tweet in tweet_list:
                    del tweets[tweet.id]

    tweets_not_location.update({0: tweets.values()})
    return tweets_not_location


# Create the id of the time slots of length time_slot_size from start_time to stop_time
def create_time_slots_id(start_time, stop_time, time_slot_size):
    i = 0
    time_slots_id = {}
    current_start_time = start_time
    while current_start_time < stop_time:
        current_stop_time = current_start_time + timedelta(minutes=time_slot_size)
        if current_stop_time <= stop_time:
            range = {'start': current_start_time, 'stop': current_stop_time}
        else:
            range = {'start': current_start_time, 'stop': stop_time}
        time_slots_id[i] = range
        current_start_time = current_stop_time
        i += 1
    return time_slots_id


# The input is the initial point of the square, the distance of the next point and the angle between the two points
def calculate_point_distance(lat, lon, distance, bearing):
    origin = geopy.Point(lat, lon)
    destination = VincentyDistance(meters=distance).destination(origin, bearing)
    return {'lon': destination.longitude, 'lat': destination.latitude}


def create_squares_id(area_northwest_point, area_x_side, area_y_side, square_side):
    cols_coordinates_id = {}
    rows_coordinates_id = {}
    # Starting from the north-west point of the rectangle, calculate the next point's coordinates moving
    # right for square_side length and assign an incremental id to the range
    iteration_point = area_northwest_point
    for i in xrange(area_x_side / square_side):
        next_point = calculate_point_distance(iteration_point['lat'], iteration_point['lon'], square_side, 90)
        points = {'sx_limit': iteration_point['lon'], 'dx_limit': next_point['lon']}
        cols_coordinates_id[i] = points
        iteration_point = next_point
    # Starting from the north-west point of the rectangle, calculate the next point's coordinates moving
    # down for square_side length and assign an incremental id to the range
    iteration_point = area_northwest_point
    for i in xrange(area_y_side / square_side):
        next_point = calculate_point_distance(iteration_point['lat'], iteration_point['lon'], square_side, 180)
        points = {'up_limit': iteration_point['lat'], 'dw_limit': next_point['lat']}
        rows_coordinates_id[i] = points
        iteration_point = next_point

    return cols_coordinates_id, rows_coordinates_id


# Assign each tweet to a square within a city
def group_tweets_by_squares(tweets, cols_coordinates_id, rows_coordinates_id, mask):
    tweets_sq = {}
    for id, tweet in tweets.iteritems():
        row_id_square = -1
        col_id_square = -1

        # Check in which range of latitude and longitude the tweet is located and store the ids
        for col_id, limit in cols_coordinates_id.iteritems():
            if limit['sx_limit'] <= tweet.position['lon'] < limit['dx_limit']:
                col_id_square = col_id
                break
        for row_id, limit in rows_coordinates_id.iteritems():
            if limit['dw_limit'] <= tweet.position['lat'] < limit['up_limit']:
                row_id_square = row_id
                break
        # Creation of the id of the square. Apply a mask to row index to keep only the first 16 bit and shift them. Then, extract
        # also the first 16 bit from the column index and then combine them with a OR. The result is the unique id of the square.
        # Then, assign the tweet to the square
        if row_id_square >= 0 and col_id_square >= 0:
            index_r = (row_id_square & mask) << 16
            index_c = col_id_square & mask
            sq_id = index_r | index_c
            if sq_id in tweets_sq.keys():
                tweets_sq[sq_id].append(tweet)
            else:
                tweets_sq[sq_id] = [tweet]
    return tweets_sq


# For each square of the city, divide it in time slot and assign the tweets
def group_tweets_by_timeslot(tweets_sq, time_slots_id):
    tweets_sq_ts = {}
    for sq_id, tweet_list in tweets_sq.items():
        for tweet in tweet_list:
            for ts_id, interval in time_slots_id.items():
                if interval['start'] <= tweet.time < interval['stop']:
                    if sq_id in tweets_sq_ts.keys():
                        if ts_id in tweets_sq_ts[sq_id].keys():
                            tweets_sq_ts[sq_id][ts_id].append(tweet)
                            continue
                        else:
                            tweets_sq_ts[sq_id].update({ts_id: [tweet]})
                            continue
                    else:
                        tweets_sq_ts.update({sq_id: {ts_id: [tweet]}})
                        continue
    return tweets_sq_ts


# Group the tweets for which a prediction will be given into the timeslots
def group_tweets_to_predict_by_timeslot(tweets, time_slots_id):
    print "Grouping tweets without coordinates in time slots...\n"
    tweets_ts = {}
    for tweet in tweets.values():
        for ts_id, interval in time_slots_id.items():
            if interval['start'] <= tweet.time < interval['stop']:
                if ts_id in tweets_ts.keys():
                    tweets_ts[ts_id].append(tweet)
                    continue
                else:
                    tweets_ts[ts_id] = [tweet]
                    continue
    return tweets_ts


def calculate_differential_heatmap(conf, start_ts, train, sq_counter_curr):
    sq_counter_prev = {}
    if start_ts >= conf.n_ts_per_day or conf.thresholds_structure:
        sq_counter_prev = sq_counter_curr
        sq_counter_curr = {}
        for loc_id, squares in train.items():
            sq_counter_curr.update({loc_id: {}})
            for sq_id, tweets_list in squares.items():
                sq_counter_curr[loc_id].update({sq_id: len(tweets_list)})
    elif conf.n_ts_per_day - 2 <= start_ts < conf.n_ts_per_day:
        sq_counter_curr = {}
        for loc_id, squares in train.items():
            sq_counter_curr.update({loc_id: {}})
            for sq_id, tweets_list in squares.items():
                sq_counter_curr[loc_id].update({sq_id: len(tweets_list)})

    return sq_counter_curr, sq_counter_prev


def merge_results(conf):
    merged_results = {}
    folder_path = os.path.join(conf.dir_path, 'Results/')
    for tweet_file in glob.glob(folder_path + "*"):
        with open(tweet_file, 'r') as file_res:
            i = 0
            for res in file_res:
                if i == 0:
                    i += 1
                    continue
                info = res.split('\t')

                if info[1] == 'no':
                    continue
                else:
                    if int(info[0]) not in merged_results.keys():
                        merged_results.update({int(info[0]): {}})
                    if int(info[1]) not in merged_results[int(info[0])].keys():
                        merged_results[int(info[0])].update({int(info[1]): {'prec': []}})
                        merged_results[int(info[0])][int(info[1])].update({'rec': []})
                        merged_results[int(info[0])][int(info[1])].update({'f1': []})

                    merged_results[int(info[0])][int(info[1])]['prec'].append(float(info[2]))
                    merged_results[int(info[0])][int(info[1])]['rec'].append(float(info[3]))
                    merged_results[int(info[0])][int(info[1])]['f1'].append(float(info[4]))

    for k in merged_results.keys():
        for j in merged_results[k].keys():
            prec = np.mean(merged_results[k][j]['prec'])
            rec = np.mean(merged_results[k][j]['rec'])
            f1 = np.mean(merged_results[k][j]['f1'])

            merged_results[k][j]['prec'] = None
            merged_results[k][j]['rec'] = None
            merged_results[k][j]['f1'] = None

            merged_results[k][j]['prec'] = prec
            merged_results[k][j]['rec'] = rec
            merged_results[k][j]['f1'] = f1

    return merged_results


################ Methods for online computation ############################################
############################################################################################

def retrieve_tweets_online(conf, file_list):
    count_good_tweets = 0
    count_bad_tweets = 0
    count_pred_tweets = 0
    pred_tweets = {}
    new_tweets = {}
    now = datetime.now() + timedelta(conf.timeslot_size)
    start_datetime = datetime.strptime(now.strftime('%a %b %d %H:%M:%S +0000 %Y'), '%a %b %d %H:%M:%S +0000 %Y')
    for tweet_file_name in file_list:
        print("Reading file {0}".format(tweet_file_name))
        with open(tweet_file_name, 'r') as tweets_file:
            for tweet in tweets_file:
                if tweet == '\n':
                    continue
                else:
                    try:
                        tweet_json = json.loads(tweet, object_hook=_decode_dict)

                        time = parse_tweet_time(tweet_json['created_at'])  # time is a datetime type
                        if time < start_datetime or not start_datetime:
                            start_datetime = time
                        id_str = tweet_json['id_str']
                        lang = tweet_json['lang']
                        # if not check_tweet_start_stop_time(time, conf.start_dt, conf.stop_dt):
                        #    continue
                        text = parse_tweet_text(conf, tweet_json['text'])  # text is parsed to keep only words
                        try:
                            position = parse_tweet_position(
                                tweet_json['coordinates']['coordinates'])  # position is a dict composed of lat and lon
                            count_good_tweets += 1
                            new_tweets.update({id_str: Tweet(id_str, lang, time, text, position)})
                        except:
                            position = None
                            pred_tweets.update({id_str: Tweet(id_str, lang, time, text, position)})
                            count_pred_tweets += 1
                    except:
                        count_bad_tweets += 1
                        continue
    return new_tweets, pred_tweets, start_datetime


def create_time_slots_id_online(start_time, actual_ts, conf):
    time_slots_id = {}
    current_start_time = start_time
    for i in range(actual_ts, actual_ts + conf.window):
        current_stop_time = current_start_time + timedelta(minutes=conf.timeslot_size)
        datetime_range = {'start': current_start_time, 'stop': current_stop_time}
        time_slots_id.update({i: datetime_range})
        current_start_time = current_stop_time
    return time_slots_id


################ Methods for saving information into files #################################
############################################################################################


def save_location_coordinates(conf, location, coordinates):
    with open(conf.dir_path + "/Visualization_data/Squares_data/" + str(location) + "/columns.txt",
              'w') as columns_file:
        for col_id, col_coordinates in coordinates[0].items():
            columns_file.write(str(col_id) + "\t" + str(col_coordinates['sx_limit']) + "\t" + str(
                col_coordinates['dx_limit']) + "\n")
    with open(conf.dir_path + "/Visualization_data/Squares_data/" + str(location) + "/rows.txt",
              'w') as rows_file:
        for row_id, row_coordinates in coordinates[1].items():
            rows_file.write(str(row_id) + "\t" + str(row_coordinates['up_limit']) + "\t" + str(
                row_coordinates['dw_limit']) + "\n")


def save_squares_time_series(tweets, time_slots_id, conf):
    for loc_id, sq_list in tweets.items():
        for sq_id, ts_list in sq_list.items():
            occurrences = [0 for y in time_slots_id.keys()]
            for ts_id, tweet_list in ts_list.items():
                occurrences[ts_id] = len(tweet_list)

            x = np.arange(len(occurrences))
            plt.figure(figsize=(10, 7))
            ax = plt.subplot(111)
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.get_xaxis().tick_bottom()
            ax.get_yaxis().tick_left()

            plt.ylim(0, max(occurrences) + 5)
            plt.xlim(0, len(occurrences) + 5)
            plt.yticks(fontsize=18)
            plt.xticks(fontsize=18)
            # plt.xticks(x, [str(h) for h in range(0,len(occurrences),20)], fontsize=15)
            plt.xlabel("Timeslots ID", size=18)
            plt.ylabel("N. of tweets", size=18)
            plt.plot(x, occurrences, lw=2, color='b', linestyle='-')
            pt = mlines.Line2D([], [], color=(123 / 255, 223 / 255, 175 / 255),
                               label=loc_id + ", square #" + str(sq_id), linestyle='-')
            plt.legend(handles=[pt], prop={'size': 18})
            plt.savefig(
                conf.dir_path + "/Visualization_data/Squares_data/" + loc_id + "/Time_series/" + str(sq_id) + ".png")
            plt.clf()
            plt.close()


def save_results(conf, cycle, results):
    with open(conf.dir_path + '/Results/run_' + str(cycle) + '.txt', 'w') as f_results:
        f_results.write("percentage\ttreshold\tprecision\trecall\tf1\n")
        for p in sorted(list(results.keys())):
            for t in sorted(list(results[p].keys())):
                avg_precision = np.mean(results[p][t]['precision'])
                avg_recall = np.mean(results[p][t]['recall'])
                avg_f1 = np.mean(results[p][t]['f1'])

                f_results.write(
                    str(p) + "\t" + str(t) + "\t" + str(avg_precision) + "\t" + str(avg_recall) + "\t" + str(
                        avg_f1) + "\n")


def save_test_data(test, prediction, location_squares, date_time, conf):
    with open(conf.dir_path + "/Visualization_data/Points_data/Test/test_" + date_time.__str__() + ".csv",
              "w") as test_file:
        test_file.write("lat,lon\n")
        for l1_id, l2_list in test.iteritems():
            for l2_id, tweet_list in l2_list.iteritems():
                for tweet in tweet_list:
                    if tweet.id in prediction.keys():
                        info = prediction[tweet.id].split(" ")
                        if info[0] != '':
                            if l1_id == info[0] and l2_id == int(info[1]):
                                sq_id = int(info[1])
                                index_c = sq_id & conf.mask
                                index_r = sq_id >> 16
                                lon = random.uniform(location_squares[l1_id][0][index_c]['sx_limit'],
                                                     location_squares[l1_id][0][index_c]['dx_limit'])
                                lat = random.uniform(location_squares[l1_id][1][index_r]['dw_limit'],
                                                     location_squares[l1_id][1][index_r]['up_limit'])
                                test_file.write(str(lat) + "," + str(lon) + "\n")


def save_train_data(train, date_time, conf):
    with open(conf.dir_path + "/Visualization_data/Points_data/Train/train_" + date_time.__str__() + ".csv",
              "w") as train_file:
        train_file.write("lat,lon\n")
        for t_1, t2_list in train.iteritems():
            for t_2, tweets_b in t2_list.iteritems():
                for tweet_a in tweets_b:
                    train_file.write(str(tweet_a.position['lat']) + "," + str(tweet_a.position['lon']) + "\n")


def save_square_train_test_counter(train, test, prediction, date_time, conf):
    for loc_id, sq_list in train.items():
        for sq_id, tweets in sq_list.items():
            n_train_tweets = len(tweets)
            with open(
                                                                            conf.dir_path + "/Visualization_data/Squares_data/" + loc_id + "/" + date_time.__str__() + "/" + str(
                                    sq_id) + ".txt", 'a') as sq_file:
                sq_file.write("train={0}\n".format(n_train_tweets))
    for l1_id, l2_list in test.iteritems():
        for l2_id, tweet_list in l2_list.iteritems():
            n_test_tweets = 0
            for tweet in tweet_list:
                if tweet.id in prediction.keys():
                    info = prediction[tweet.id].split(" ")
                    if l1_id == info[0] and l2_id == int(info[1]):
                        n_test_tweets += 1
            with open(
                                                                            conf.dir_path + "/Visualization_data/Squares_data/" + l1_id + "/" + date_time.__str__() + "/" + str(
                                    l2_id) + ".txt", 'a') as sq_file:
                sq_file.write("test={0}\n".format(n_test_tweets))


def save_square_keywords(loc_id, documents, date_time, conf):
    for sq_id, document in documents.items():
        sorted_words = sorted(document.items(), key=operator.itemgetter(1))
        words = ""
        for word in sorted_words[:6]:
            words += str(word[0]) + " "
        with open(conf.dir_path + "/Visualization_data/Squares_data/" + loc_id + "/" + date_time.__str__() + "/" + str(
                sq_id) + ".txt", 'a') as sq_file:
            sq_file.write("keywords={0}\n".format(words))


def save_square_differential(cur_counter, prev_counter, date_time, conf):
    diff_value = 0.0
    if not prev_counter:
        return
    for loc_id in cur_counter.keys():
        squares_list = set().union(cur_counter[loc_id].keys(), prev_counter[loc_id].keys())
        for sq_id in squares_list:
            if sq_id in cur_counter[loc_id].keys() and sq_id in prev_counter[loc_id].keys():
                diff_value = 100.0 * (
                    float(cur_counter[loc_id][sq_id] - prev_counter[loc_id][sq_id]) / float(
                        prev_counter[loc_id][sq_id]))
            elif sq_id in prev_counter[loc_id].keys() and not sq_id in cur_counter[loc_id].keys():
                diff_value = -100.0
            elif sq_id in cur_counter[loc_id].keys() and not sq_id in prev_counter[loc_id].keys():
                diff_value = 100.0

            with open(
                                                                            conf.dir_path + "/Visualization_data/Squares_data/" + loc_id + "/" + date_time.__str__() + "/" + str(
                                    sq_id) + ".txt", 'a') as sq_file:
                sq_file.write("differential={0}\n".format(round(diff_value, 2)))


def save_tweets_not_location(tweets, date_time, start_ts, conf):
    with open(conf.dir_path + "/Visualization_data/Points_data/Train/train_" + date_time.__str__() + ".csv",
              "a") as train_file:
        ts = tweets[0]
        for ts_id in ts.keys():
            if start_ts <= ts_id < start_ts + conf.window:
                for tweet in ts[ts_id]:
                    train_file.write(str(tweet.position['lat']) + "," + str(tweet.position['lon']) + "\n")


def save_thresholds(conf, thresholds):
    file_handler = open(conf.dir_path + "/thresholds.data", "wb")
    pickle.dump(thresholds, file_handler)
    file_handler.close()


################ Methods for printing information ##########################################
############################################################################################


def print_window_progress(q_size, w_size):
    red_bar = ""
    for b in range(q_size):
        red_bar = red_bar + "|"
    white_bar = "-" * (w_size - (q_size))
    sys.stdout.write(
        '\rWait for the window to be full\t' + red_bar + white_bar + "  " + str(q_size) + "/" + str(w_size))
    sys.stdout.flush()


def print_time_slots_progress(time_slot):
    sys.stdout.write('\nStart timeslot of the window: ' + str(time_slot) + '\n\n')
    sys.stdout.flush()


################ Methods to predict location of tweets #####################################
############################################################################################


def save_prediction(tweets, top, location_squares, date_time, conf):
    with open(conf.dir_path + "/Predictions/" + date_time.__str__() + ".json", "w") as pred_file:
        for id, place in top.items():
            info = place.split(" ")
            if info[0] != '':
                loc_id = info[0]
                sq_id = int(info[1])
                index_c = sq_id & conf.mask
                index_r = sq_id >> 16
                lon = random.uniform(location_squares[loc_id][0][index_c]['sx_limit'],
                                     location_squares[loc_id][0][index_c]['dx_limit'])
                lat = random.uniform(location_squares[loc_id][1][index_r]['dw_limit'],
                                     location_squares[loc_id][1][index_r]['up_limit'])

                for tweet in tweets:
                    if tweet.id == id:
                        tweet.position = {}
                        tweet.position.update({'lat': lat})
                        tweet.position.update({'lon': lon})

                        pred_file.write(tweet.toJson() + "\n")


def find_tweet_position(tweets, key):
    for l2 in tweets.values():
        for tweet_list in l2.values():
            for tweet in tweet_list:
                if tweet.id == key:
                    return (tweet.position['lat'], tweet.position['lon'])


def assign_tweet_location(info, location_squares, conf):
    sq_id = int(info[1])
    index_c = sq_id & conf.mask
    index_r = sq_id >> 16
    lon = (location_squares[info[0]][0][index_c]['sx_limit'] + location_squares[info[0]][0][index_c]['dx_limit']) / 2
    lat = (location_squares[info[0]][1][index_r]['dw_limit'] + location_squares[info[0]][1][index_r]['up_limit']) / 2
    return (lat, lon)
